package org.apidesign.apifest08.test;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;
import junit.framework.TestCase;
import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.CurrencyValue;
import org.apidesign.apifest08.currency.ExchangeRateValue;

/** The exchange rates are not always the same. They are changing. However
 * as in order to predict the future, one needs to understand own past. That is
 * why it is important to know the exchange rate as it was at any time during
 * the past.
 * <p>
 * Today's quest is to enhance the convertor API to deal with dates.
 * One shall be able to convert a currency at any date. Each currencies rate shall
 * be associated with a range between two Date objects. In order
 * to keep compatibility with old API that knew nothing about dates, the
 * rates associated then are applicable "for eternity". Any use of existing
 * convert methods that do not accept a Date argument, uses the current
 * System.currentTimeMillis() as default date.
 */
public class Task4Test extends TestCase {
    public Task4Test(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
    }

    @Override
    protected void tearDown() throws Exception {
    }

    // Backward compatibly enhance your existing API to support following
    // usecases:
    //

    /** Takes a convertor with any rates associated and creates new convertor
     * that returns the same values as the old one for time between from to till.
     * Otherwise it returns no results. This is just a helper method that
     * shall call some real one in the API.
     * 
     * @param old existing convertor
     * @param from initial date (inclusive)
     * @param till final date (exclusive)
     * @return new convertor
     */
    public static <AmountType, IdentifierType> Convertor<AmountType, IdentifierType> limitTo(Convertor<AmountType, IdentifierType> old, Date from, Date till) {
        return old.limitConvertor(from, till);
    }


    public void testCompositionOfLimitedConvertors() throws Exception {
        Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
        cal.clear();
        cal.set(2008, 9, 1, 0, 0);
        Date d1 = cal.getTime(); // 2008-10-01 0:00 GMT
        cal.set(2008, 9, 2, 0, 0);
        Date d2 = cal.getTime(); // 2008-10-02 0:00 GMT
        cal.set(2008, 9, 3, 0, 0);
        Date d3 = cal.getTime(); // 2008-10-03 0:00 GMT
        
        Convertor<Double, String> c = Task2Test.merge(
            limitTo(Task1Test.createCZKtoUSD(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        CurrencyValue<Double, String> result;
        // convert $5 to CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("CZK", CurrencyValue.getCurrencyValue(5d, "USD"));
            fail("Should not convert");
        } catch (Exception e) {
        }
        
        // convert $8 to CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("CZK", CurrencyValue.getCurrencyValue(8d, "USD"));
            fail("Should not convert");
        } catch (Exception e) {
        }

        // convert 1003CZK to USD using c:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("USD", CurrencyValue.getCurrencyValue(1003d, "CZK"));
            fail("Should not convert");
        } catch (Exception e) {
        }

        // convert 16CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("SKK", CurrencyValue.getCurrencyValue(16d, "CZK"));
            fail("Should not convert");
        } catch (Exception e) {
        }

        // convert 500SKK to CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("CZK", CurrencyValue.getCurrencyValue(500d, "SKK"));
            fail("Should not convert");
        } catch (Exception e) {
        }

        // convert $5 to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 85 CZK");
        cal.set(2008, 9, 1, 6, 0);
        result = c.convert("CZK", CurrencyValue.getCurrencyValue(5d, "USD"), cal.getTime());
        assertEquals(CurrencyValue.getCurrencyValue(85d, "CZK"), result);

        // convert $8 to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 136 CZK");
        cal.set(2008, 9, 1, 6, 0);
        result = c.convert("CZK", CurrencyValue.getCurrencyValue(8d, "USD"), cal.getTime());
        assertEquals(CurrencyValue.getCurrencyValue(136d, "CZK"), result);

        // convert 1003CZK to USD using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 59 USD");
        cal.set(2008, 9, 1, 6, 0);
        result = c.convert("USD", CurrencyValue.getCurrencyValue(1003d, "CZK"), cal.getTime());
        assertEquals(CurrencyValue.getCurrencyValue(59d, "USD"), result);

        // convert 16CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 20 SKK");
        cal.set(2008, 9, 2, 9, 0);
        result = c.convert("SKK", CurrencyValue.getCurrencyValue(16d, "CZK"), cal.getTime());
        assertEquals(CurrencyValue.getCurrencyValue(20d, "SKK"), result);

        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 400 CZK");
        cal.set(2008, 9, 2, 9, 0);
        result = c.convert("CZK", CurrencyValue.getCurrencyValue(500d, "SKK"), cal.getTime());
        assertEquals(CurrencyValue.getCurrencyValue(400d, "CZK"), result);

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        // cannot convert as no rate is applicable to current date
        cal.set(2008, 9, 1, 6, 0);
        try {
            result = c.convert("CZK", CurrencyValue.getCurrencyValue(500d, "SKK"), cal.getTime());
            fail("Should not convert");
        } catch (Exception e) {
        }
    }

    /** Create convertor that understands two currencies, CZK and
     *  SKK. Make 100 SKK == 90 CZK.
     *
     * @return prepared convertor ready for converting SKK to CZK and CZK to SKK
     */
    public static Convertor<Double, String> createSKKtoCZK2() {
        return Convertor.getConvertorDoubleString(
                ExchangeRateValue.getExchangeRate(
                    CurrencyValue.getCurrencyValue(100d, "SKK"),
                    CurrencyValue.getCurrencyValue(90d, "CZK")));
    }

    public void testDateConvetorWithTwoDifferentRates() throws Exception {
        Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
        cal.clear();
        cal.set(2008, 9, 1, 0, 0);
        Date d1 = cal.getTime(); // 2008-10-01 0:00 GMT
        cal.set(2008, 9, 2, 0, 0);
        Date d2 = cal.getTime(); // 2008-10-02 0:00 GMT
        cal.set(2008, 9, 3, 0, 0);
        Date d3 = cal.getTime(); // 2008-10-03 0:00 GMT

        Convertor<Double, String> c = Task2Test.merge(
            limitTo(createSKKtoCZK2(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        CurrencyValue<Double, String> result;
        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 400 CZK");
        cal.set(2008, 9, 2, 9, 0);
        result = c.convert("CZK", CurrencyValue.getCurrencyValue(500d, "SKK"), cal.getTime());
        assertEquals(CurrencyValue.getCurrencyValue(400d, "CZK"), result);

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 450 CZK");
        cal.set(2008, 9, 1, 6, 0);
        result = c.convert("CZK", CurrencyValue.getCurrencyValue(500d, "SKK"), cal.getTime());
        assertEquals(CurrencyValue.getCurrencyValue(450d, "CZK"), result);
    }

}
