package org.apidesign.apifest08.test;

import java.math.BigDecimal;
import java.text.SimpleDateFormat;
import java.util.Date;
import junit.framework.TestCase;
import org.apidesign.apifest08.currency.ConversionNotSupportedException;
import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.ConvertorCurrency;
import org.apidesign.apifest08.currency.ExchangeRateProvider;

/** The exchange rates are not always the same. They are changing. However
 * as in order to predict the future, one needs to understand own past. That is
 * why it is important to know the exchange rate as it was at any time during
 * the past.
 * <p>
 * Today's quest is to enhance the convertor API to deal with dates.
 * One shall be able to convert a currency at any date. Each currencies rate shall
 * be associated with a range between two Date objects. In order
 * to keep compatibility with old API that knew nothing about dates, the
 * rates associated then are applicable "for eternity". Any use of existing
 * convert methods that do not accept a Date argument, uses the current
 * System.currentTimeMillis() as default date.
 */
public class Task4Test extends TestCase {
    private static ConvertorCurrency CZK = ConvertorCurrency.getInstance("CZK");
    private static ConvertorCurrency SKK = ConvertorCurrency.getInstance("SKK");
    private static ConvertorCurrency USD = ConvertorCurrency.getInstance("USD");
    private SimpleDateFormat df;
    
    
    public Task4Test(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
        super.setUp();
        df = new SimpleDateFormat("yyyy-MM-dd HH:mm zzzz");
    }

    @Override
    protected void tearDown() throws Exception {
        super.tearDown();
        df = null;
    }

    // Backward compatibly enhance your existing API to support following
    // usecases:
    //

    /** Takes a convertor with any rates associated and creates new convertor
     * that returns the same values as the old one for time between from to till.
     * Otherwise it returns no results. This is just a helper method that
     * shall call some real one in the API.
     * 
     * @param old existing convertor
     * @param from initial date (inclusive)
     * @param till final date (exclusive)
     * @return new convertor
     */
    public static Convertor limitTo(Convertor old, Date from, Date till) {
        Convertor c = Convertor.createConvertorAsMerge(new Convertor[]{old});
        c.limitAllowedDates(from, till);
        return c;
    }


    public void testCompositionOfLimitedConvertors() throws Exception {
        Date d1 = df.parse("2008-10-01 0:00 GMT");
        Date d2 = df.parse("2008-10-02 0:00 GMT");
        Date d3 = df.parse("2008-10-03 0:00 GMT");
        
        Convertor c = Task2Test.merge(
            limitTo(Task1Test.createCZKtoUSD(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        // convert $5 to CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
          c.convert(USD, CZK, new BigDecimal(5));
          fail();        
        } catch (ConversionNotSupportedException e) {
            //exception expected
        }

        // convert $8 to CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
          c.convert(USD, CZK, new BigDecimal(8));
          fail();        
        } catch (ConversionNotSupportedException e) {
            //exception expected
        }
        

        // convert 1003CZK to USD using c:
        // cannot convert as no rate is applicable to current date
        try {
          c.convert(CZK, USD, new BigDecimal(1003));
          fail();        
        } catch (ConversionNotSupportedException e) {
            //exception expected
        }

        // convert 16CZK using c:
        // cannot convert as no rate is applicable to current date
        // ???
        try {
          c.convert(CZK, USD, new BigDecimal(16));
          fail();        
        } catch (ConversionNotSupportedException e) {
            //exception expected
        }

        // convert 500SKK to CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
          c.convert(SKK, CZK, new BigDecimal(500));
          fail();        
        } catch (ConversionNotSupportedException e) {
            //exception expected
        }
        

        // convert $5 to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 85 CZK");
        assertEquals("Result is 85 CZK", new BigDecimal("85.00"), c.convertWithReversibleRates(USD, CZK , new BigDecimal("5"),df.parse("2008-10-01 6:00 GMT")).getConverted());        

        // convert $8 to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 136 CZK");
        assertEquals("Result is 136 CZK", new BigDecimal("136.00"), c.convertWithReversibleRates(USD, CZK , new BigDecimal("8"),df.parse("2008-10-01 6:00 GMT")).getConverted());        

        // convert 1003CZK to USD using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 59 USD");
        assertEquals("Result is 59 USD", new BigDecimal("59.00"), c.convertWithReversibleRates(CZK, USD , new BigDecimal("1003"),df.parse("2008-10-01 6:00 GMT")).getConverted());        

        // convert 16CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 20 SKK");
        assertEquals("Result is 20 SKK", new BigDecimal("20.00"), c.convertWithReversibleRates(CZK, SKK , new BigDecimal("16"),df.parse("2008-10-02 9:00 GMT")).getConverted());        

        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 400 CZK");
        assertEquals("Result is 400 SKK", new BigDecimal("400.00"), c.convertWithReversibleRates(SKK, CZK , new BigDecimal("500"),df.parse("2008-10-02 9:00 GMT")).getConverted());        

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        // cannot convert as no rate is applicable to current date
        try {
          c.convertWithReversibleRates(SKK, CZK , new BigDecimal("500"),df.parse("2008-10-01 6:00 GMT")).getConverted();        
          fail();        
        } catch (ConversionNotSupportedException e) {
            //exception expected
        }

    }

    /** Create convertor that understands two currencies, CZK and
     *  SKK. Make 100 SKK == 90 CZK.
     *
     * @return prepared convertor ready for converting SKK to CZK and CZK to SKK
     */
    public static Convertor createSKKtoCZK2() {
        ExchangeRateProvider exchangeRateProvider = ExchangeRateProvider.createExchangeRateProvider();
        exchangeRateProvider.addFixedCurencyRate(SKK, new BigDecimal("100.00"), CZK, new BigDecimal("90.00"));
        Convertor c  = Convertor.createConvertor(exchangeRateProvider);
        return c;
    }

    public void testDateConvetorWithTwoDifferentRates() throws Exception {

        Date d1 = df.parse("2008-10-01 0:00 GMT");
        Date d2 = df.parse("2008-10-02 0:00 GMT");
        Date d3 = df.parse("2008-10-03 0:00 GMT");

        Convertor c = Task2Test.merge(
            limitTo(createSKKtoCZK2(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        assertEquals("Result is 400 CZK", new BigDecimal("400.00"), c.convertWithReversibleRates(SKK, CZK , new BigDecimal("500"),df.parse("2008-10-02 9:00 GMT")).getConverted());        

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        assertEquals("Result is 450 CZK", new BigDecimal("450.00"), c.convertWithReversibleRates(SKK, CZK , new BigDecimal("500"),df.parse("2008-10-01 9:00 GMT")).getConverted());        
    }

}
